@@ -23,12 +23,13 @@ import libusb1
 from ctypes import byref, create_string_buffer, c_int, sizeof, POINTER, \
     create_unicode_buffer, c_wchar, cast, c_uint16, c_ubyte, string_at, \
     c_void_p, cdll
-from cStringIO import StringIO
 import sys
 import threading
 from ctypes.util import find_library
+import warnings
+import weakref
 
-__all__ = ['LibUSBContext', 'USBDeviceHandle', 'USBDevice',
+__all__ = ['USBContext', 'USBDeviceHandle', 'USBDevice',
     'USBPoller', 'USBTransfer', 'USBTransferHelper', 'EVENT_CALLBACK_SET']
 
 if sys.version_info[:2] >= (2, 6):
@@ -50,6 +51,20 @@ else:
     _free = getattr(cdll, __libc_name).free
 del __libc_name
 
+try:
+    WeakSet = weakref.WeakSet
+except AttributeError:
+    # Python < 2.7: tiny wrapper around WeakKeyDictionary
+    class WeakSet(object):
+        def __init__(self):
+            self.__dict = weakref.WeakKeyDictionary()
+
+        def add(self, item):
+            self.__dict[item] = None
+
+        def pop(self):
+            return self.__dict.popitem()[0]
+
 # Default string length
 # From a comment in libusb-1.0: "Some devices choke on size > 255"
 STRING_LENGTH = 255
@@ -74,6 +89,9 @@ def create_binary_buffer(string_or_len):
         result = create_string_buffer(string_or_len, len(string_or_len))
     return result
 
+class DoomedTransferError(Exception):
+    pass
+
 class USBTransfer(object):
     """
     USB asynchronous transfer control & data.
@@ -81,6 +99,11 @@ class USBTransfer(object):
     All modification methods will raise if called on a submitted transfer.
     Methods noted as "should not be called on a submitted transfer" will not
     prevent you from reading, but returned value is unspecified.
+
+    Note on user_data: because of pypy's current ctype restrictions, user_data
+    is not provided to C level, but is managed purely in python. It should
+    change nothing for you, unless you are looking at underlying C transfer
+    structure - which you should never have to.
     """
     # Prevent garbage collector from freeing the free function before our
     # instances, as we need it to property destruct them.
@@ -93,8 +116,11 @@ class USBTransfer(object):
     __submitted = False
     __callback = None
     __ctypesCallbackWrapper = None
+    __doomed = False
+    __user_data = None
+    __transfer_buffer = None
 
-    def __init__(self, handle, iso_packets=0):
+    def __init__(self, handle, iso_packets, before_submit, after_completion):
         """
         You should not instanciate this class directly.
         Call "getTransfer" method on an USBDeviceHandle instance to get
@@ -104,6 +130,8 @@ class USBTransfer(object):
             raise ValueError('Cannot request a negative number of iso '
                 'packets.')
         self.__handle = handle
+        self.__before_submit = before_submit
+        self.__after_completion = after_completion
         self.__num_iso_packets = iso_packets
         result = libusb1.libusb_alloc_transfer(iso_packets)
         if not result:
@@ -114,21 +142,34 @@ class USBTransfer(object):
 
     def close(self):
         """
-        Stop using this transfer.
-        This removes some references to other python objects, to help garbage
-        collection.
+        Break reference cycles to allow instance to be garbage-collected.
         Raises if called on a submitted transfer.
-        This does not prevent future reuse of instance (calling one of
-        "setControl", "setBulk", "setInterrupt" or "setIsochronous" methods
-        will initialize it properly again), just makes it ready to be
-        garbage-collected.
-        It is not mandatory to call it either, if you have no problems with
-        garbage collection.
         """
         if self.__submitted:
             raise ValueError('Cannot close a submitted transfer')
+        self.doom()
         self.__initialized = False
+        # Break possible external reference cycle
         self.__callback = None
+        # Break libusb_transfer reference cycles
+        self.__ctypesCallbackWrapper = None
+        # For some reason, overwriting callback is not enough to remove this
+        # reference cycle - though sometimes it works:
+        #   self -> self.__dict__ -> libusb_transfer -> dict[x] -> dict[x] ->
+        #   CThunkObject -> __callbackWrapper -> self
+        # So free transfer altogether.
+        if self.__transfer is not None:
+            self.__libusb_free_transfer(self.__transfer)
+            self.__transfer = None
+        # Break USBDeviceHandle reference cycle
+        self.__before_submit = None
+        self.__after_completion = None
+
+    def doom(self):
+        """
+        Prevent transfer from being submitted again.
+        """
+        self.__doomed = True
 
     def __del__(self):
         if self.__transfer is not None:
@@ -156,9 +197,12 @@ class USBTransfer(object):
         fired (ie, mark transfer as not submitted upon call).
         """
         self.__submitted = False
+        self.__after_completion(self)
         callback = self.__callback
         if callback is not None:
             callback(self)
+        if self.__doomed:
+            self.close()
 
     def setCallback(self, callback):
         """
@@ -190,6 +234,8 @@ class USBTransfer(object):
         """
         if self.__submitted:
             raise ValueError('Cannot alter a submitted transfer')
+        if self.__doomed:
+            raise DoomedTransferError('Cannot reuse a doomed transfer')
         if isinstance(buffer_or_len, basestring):
             length = len(buffer_or_len)
             string_buffer = create_binary_buffer(
@@ -199,10 +245,12 @@ class USBTransfer(object):
             string_buffer = create_binary_buffer(length + \
                 libusb1.LIBUSB_CONTROL_SETUP_SIZE)
         self.__initialized = False
+        self.__transfer_buffer = string_buffer
+        self.__user_data = user_data
         libusb1.libusb_fill_control_setup(string_buffer, request_type,
             request, value, index, length)
         libusb1.libusb_fill_control_transfer(self.__transfer, self.__handle,
-            string_buffer, self.__ctypesCallbackWrapper, user_data, timeout)
+            string_buffer, self.__ctypesCallbackWrapper, None, timeout)
         self.__callback = callback
         self.__initialized = True
 
@@ -223,11 +271,15 @@ class USBTransfer(object):
         """
         if self.__submitted:
             raise ValueError('Cannot alter a submitted transfer')
+        if self.__doomed:
+            raise DoomedTransferError('Cannot reuse a doomed transfer')
         string_buffer = create_binary_buffer(buffer_or_len)
         self.__initialized = False
+        self.__transfer_buffer = string_buffer
+        self.__user_data = user_data
         libusb1.libusb_fill_bulk_transfer(self.__transfer, self.__handle,
             endpoint, string_buffer, sizeof(string_buffer),
-            self.__ctypesCallbackWrapper, user_data, timeout)
+            self.__ctypesCallbackWrapper, None, timeout)
         self.__callback = callback
         self.__initialized = True
 
@@ -248,11 +300,15 @@ class USBTransfer(object):
         """
         if self.__submitted:
             raise ValueError('Cannot alter a submitted transfer')
+        if self.__doomed:
+            raise DoomedTransferError('Cannot reuse a doomed transfer')
         string_buffer = create_binary_buffer(buffer_or_len)
         self.__initialized = False
+        self.__transfer_buffer = string_buffer
+        self.__user_data = user_data
         libusb1.libusb_fill_interrupt_transfer(self.__transfer, self.__handle,
             endpoint, string_buffer,  sizeof(string_buffer),
-            self.__ctypesCallbackWrapper, user_data, timeout)
+            self.__ctypesCallbackWrapper, None, timeout)
         self.__callback = callback
         self.__initialized = True
 
@@ -283,6 +339,8 @@ class USBTransfer(object):
             raise TypeError('This transfer canot be used for isochronous I/O. '
                 'You must get another one with a non-zero iso_packets '
                 'parameter.')
+        if self.__doomed:
+            raise DoomedTransferError('Cannot reuse a doomed transfer')
         string_buffer = create_binary_buffer(buffer_or_len)
         buffer_length = sizeof(string_buffer)
         if iso_transfer_length_list is None:
@@ -299,9 +357,11 @@ class USBTransfer(object):
                     buffer_length))
         transfer_p = self.__transfer
         self.__initialized = False
+        self.__transfer_buffer = string_buffer
+        self.__user_data = user_data
         libusb1.libusb_fill_iso_transfer(transfer_p, self.__handle,
             endpoint, string_buffer, buffer_length, configured_iso_packets,
-            self.__ctypesCallbackWrapper, user_data, timeout)
+            self.__ctypesCallbackWrapper, None, timeout)
         for length, iso_packet_desc in zip(iso_transfer_length_list,
                 libusb1.get_iso_packet_list(transfer_p)):
             if length <= 0:
@@ -351,6 +411,20 @@ class USBTransfer(object):
             result = string_at(transfer.buffer, transfer.length)
         return result
 
+    def getUserData(self):
+        """
+        Retrieve user data provided on setup.
+        """
+        return self.__user_data
+
+    def setUserData(self, user_data):
+        """
+        Change user data.
+        """
+        if self.__submitted:
+            raise ValueError('Cannot alter a submitted transfer')
+        self.__user_data = user_data
+
     def getISOBufferList(self):
         """
         Get individual ISO transfer's buffer.
@@ -409,6 +483,7 @@ class USBTransfer(object):
                 sizeof(buff) != transfer.length:
             raise ValueError('To alter isochronous transfer buffer length, '
                 'use setIsochronous')
+        self.__transfer_buffer = buff
         transfer.buffer = cast(buff, c_void_p)
         transfer.length = sizeof(buff)
 
@@ -427,9 +502,13 @@ class USBTransfer(object):
         if not self.__initialized:
             raise ValueError('Cannot submit a transfer until it has been '
                 'initialized')
+        if self.__doomed:
+            raise DoomedTransferError('Cannot submit doomed transfer')
+        self.__before_submit(self)
         self.__submitted = True
         result = libusb1.libusb_submit_transfer(self.__transfer)
         if result:
+            self.__after_completion(self)
             self.__submitted = False
             raise libusb1.USBError(result)
 
@@ -439,10 +518,14 @@ class USBTransfer(object):
         Note: cancellation happens asynchronously, so you must wait for
         LIBUSB_TRANSFER_CANCELLED.
         """
+        if not self.__submitted:
+            # XXX: Workaround for a bug reported on libusb 1.0.8: calling
+            # libusb_cancel_transfer on a non-submitted transfer might
+            # trigger a segfault.
+            raise self.__USBError(self.__LIBUSB_ERROR_NOT_FOUND)
         result = self.__libusb_cancel_transfer(self.__transfer)
         if result:
             raise self.__USBError(result)
-        self.__submitted = False
 
 class USBTransferHelper(object):
     """
@@ -451,6 +534,8 @@ class USBTransferHelper(object):
     - no need to read event status to execute apropriate code, just setup
       different functions for each status code
     - just return True instead of calling submit
+    - no need to check if transfer is doomed before submitting it again,
+      DoomedTransferError is caught.
 
     Callbacks used in this class must follow the callback API described in
     USBTransfer, and are expected to return a boolean:
@@ -525,7 +610,10 @@ class USBTransferHelper(object):
         """
         if self.getEventCallback(transfer.getStatus(), self.__errorCallback)(
                 transfer):
-            transfer.submit()
+            try:
+                transfer.submit()
+            except DoomedTransferError:
+                pass
 
     def isSubmited(self):
         """
@@ -553,16 +641,8 @@ class USBPollerThread(threading.Thread):
         Warning: it will not check if another poller instance was already
         present for that context, and will replace it.
 
-        poller is a polling instance implementing the following methods:
-        - register(fd, event_flags)
-          event_flags have the same meaning as in poll API (POLLIN & POLLOUT)
-        - unregister(fd)
-        - poll(timeout)
-          timeout being a float in seconds, or None if there is no timeout.
-          Its return value must evaluate to true if there are events to handle,
-          false otherwise.
-        poller should not be used outside this class, and should not have
-        any registered file descriptor.
+        poller
+            (same as USBPoller.__init__ "poller" parameter)
 
         exc_callback (callable)
           Called with a libusb_error value as single parameter when event
@@ -649,9 +729,13 @@ class USBPoller(object):
           event_flags have the same meaning as in poll API (POLLIN & POLLOUT)
         - unregister(fd)
         - poll(timeout)
-          timeout being a float in seconds, or None if there is no timeout.
+          timeout being a float in seconds, or negative/None if there is no
+          timeout.
           It must return a list of (descriptor, event) pairs.
         Note: USBPoller is itself a valid poller.
+        Note2: select.poll uses a timeout in milliseconds, for some reason
+        (all other select.* classes use seconds for timeout), so you should
+        wrap it to convert & round/truncate timeout.
         """
         self.__context = context
         self.__poller = poller
@@ -670,7 +754,7 @@ class USBPoller(object):
         Returns a list of (descriptor, event) pairs.
         """
         next_usb_timeout = self.__context.getNextTimeout()
-        if timeout is None:
+        if timeout is None or timeout < 0:
             usb_timeout = next_usb_timeout
         elif next_usb_timeout:
             usb_timeout = min(next_usb_timeout, timeout)
@@ -721,6 +805,9 @@ class USBDeviceHandle(object):
     """
     __handle = None
     __libusb_close = libusb1.libusb_close
+    __USBError = libusb1.USBError
+    __LIBUSB_ERROR_NOT_FOUND = libusb1.LIBUSB_ERROR_NOT_FOUND
+    __set = set
 
     def __init__(self, context, handle, device):
         """
@@ -728,9 +815,20 @@ class USBDeviceHandle(object):
         Call "open" method on an USBDevice instance to get an USBDeviceHandle
         instance.
         """
-        # XXX Context parameter is just here as a hint for garbage collector:
-        # It must collect USBDeviceHandle instances before their LibUSBContext.
         self.__context = context
+        # Weak reference to transfers about this device so we can clean up
+        # before closing device.
+        self.__transfer_set = WeakSet()
+        # Strong references to inflight transfers so they do not get freed
+        # even if user drops all strong references to them. If this instance
+        # is garbage-collected, we close all transfers, so it's fine.
+        self.__inflight = inflight = set()
+        # XXX: For some reason, doing self.__inflight.{add|remove} inside
+        # getTransfer causes extra intermediate python objects for each
+        # allocated transfer. Storing them as properties solves this. Found
+        # with objgraph.
+        self.__inflight_add = inflight.add
+        self.__inflight_remove = inflight.remove
         self.__handle = handle
         self.__device = device
 
@@ -741,9 +839,42 @@ class USBDeviceHandle(object):
         """
         Close this handle. If not called explicitely, will be called by
         destructor.
+
+        This method cancels any in-flight transfer when it is called. As
+        cancellation is not immediate, this method needs to let libusb handle
+        events until transfers are actually cancelled.
+        In multi-threaded programs, this can lead to stalls. To avoid this,
+        do not close nor let GC collect a USBDeviceHandle which has in-flight
+        transfers.
         """
         handle = self.__handle
         if handle is not None:
+            # Build a strong set from weak self.__transfer_set so we can doom
+            # and close all contained transfers.
+            # Because of backward compatibility, self.__transfer_set might be a
+            # wrapper around WeakKeyDictionary. As it might be modified by gc,
+            # we must pop until there is not key left instead of iterating over
+            # it.
+            weak_transfer_set = self.__transfer_set
+            transfer_set = self.__set()
+            while True:
+                try:
+                    transfer = weak_transfer_set.pop()
+                except KeyError:
+                    break
+                transfer_set.add(transfer)
+                transfer.doom()
+            inflight = self.__inflight
+            for transfer in inflight:
+                try:
+                    transfer.cancel()
+                except self.__USBError, exception:
+                    if exception.value != self.__LIBUSB_ERROR_NOT_FOUND:
+                        raise
+            while inflight:
+                self.__context.handleEvents()
+            for transfer in transfer_set:
+                transfer.close()
             self.__libusb_close(handle)
             self.__handle = None
 
@@ -1032,7 +1163,11 @@ class USBDeviceHandle(object):
         iso_packets: the number of isochronous transfer descriptors to
           allocate.
         """
-        return USBTransfer(self.__handle, iso_packets)
+        result = USBTransfer(self.__handle, iso_packets,
+            self.__inflight_add, self.__inflight_remove,
+        )
+        self.__transfer_set.add(result)
+        return result
 
 class USBConfiguration(object):
     def __init__(self, config):
@@ -1132,16 +1267,16 @@ class USBInterfaceSetting(object):
     def iterEndpoints(self):
         endpoint_list = self.__alt_setting.endpoint
         for endpoint_num in xrange(self.getNumEndpoints()):
-            yield USBEndPoint(endpoint_list[endpoint_num])
+            yield USBEndpoint(endpoint_list[endpoint_num])
 
     def __getitem__(self, endpoint):
         if not isinstance(endpoint, int):
             raise TypeError('endpoint parameter must be an integer')
         if not (0 <= endpoint < self.getNumEndpoints()):
             raise ValueError('No such endpoint: %r' % (endpoint, ))
-        return USBEndPoint(self.__alt_setting.endpoint[endpoint])
+        return USBEndpoint(self.__alt_setting.endpoint[endpoint])
 
-class USBEndPoint(object):
+class USBEndpoint(object):
     def __init__(self, endpoint):
         if not isinstance(endpoint, libusb1.libusb_endpoint_descriptor):
             raise TypeError('Unexpected descriptor type.')
@@ -1390,7 +1525,7 @@ class USBDevice(object):
 _zero_tv = libusb1.timeval(0, 0)
 _zero_tv_p = byref(_zero_tv)
 
-class LibUSBContext(object):
+class USBContext(object):
     """
     libusb1 USB context.
 
@@ -1406,31 +1541,41 @@ class LibUSBContext(object):
     def _validContext(func):
         # Defined inside LibUSBContext so we can access "self.__*".
         def wrapper(self, *args, **kw):
-            self.__context_lock.acquire()
+            self.__context_cond.acquire()
+            self.__context_refcount += 1
+            self.__context_cond.release()
             try:
                 if self.__context_p is not None:
                     return func(self, *args, **kw)
             finally:
-                self.__context_lock.release()
+                self.__context_cond.acquire()
+                self.__context_refcount -= 1
+                if not self.__context_refcount:
+                    self.__context_cond.notifyAll()
+                self.__context_cond.release()
         return wrapper
 
     def __init__(self):
         """
         Create a new USB context.
         """
+        # Used to prevent an exit to cause a segfault if a concurrent thread
+        # is still in libusb.
+        self.__context_refcount = 0
+        self.__context_cond = threading.Condition()
         context_p = libusb1.libusb_context_p()
         result = libusb1.libusb_init(byref(context_p))
         if result:
             raise libusb1.USBError(result)
         self.__context_p = context_p
-        # Used to prevent an exit to cause a segfault if a concurent thread
-        # is still in libusb.
-        self.__context_lock = threading.RLock()
 
     def __del__(self):
-        self.exit()
+        # Avoid locking.
+        # XXX: Assumes __del__ should not normally be called while any
+        # instance's method is being executed. It seems unlikely (they hold a
+        # reference to their instance).
+        self._exit()
 
-    @_validContext
     def exit(self):
         """
         Close (destroy) this USB context.
@@ -1438,44 +1583,89 @@ class LibUSBContext(object):
         When this function has been called, methods on its instance will
         become mosty no-ops, returning None.
         """
+        self.__context_cond.acquire()
+        try:
+            while self.__context_refcount and self.__context_p is not None:
+                self.__context_cond.wait()
+            self._exit()
+        finally:
+            self.__context_cond.notifyAll()
+            self.__context_cond.release()
+
+    def _exit(self):
         context_p = self.__context_p
-        self.__libusb_exit(context_p)
-        self.__context_p = None
-        self.__added_cb = None
-        self.__removed_cb = None
+        if context_p is not None:
+            self.__libusb_exit(context_p)
+            self.__context_p = None
+            self.__added_cb = None
+            self.__removed_cb = None
 
     @_validContext
-    def getDeviceList(self):
+    def getDeviceList(self, skip_on_access_error=False):
         """
         Return a list of all USB devices currently plugged in, as USBDevice
         instances.
+
+        skip_on_access_error (bool)
+            If True, catch LIBUSB_ERROR_ACCESS errors and just skip to next
+            device.
         """
         device_p_p = libusb1.libusb_device_p_p()
         libusb_device_p = libusb1.libusb_device_p
         device_list_len = libusb1.libusb_get_device_list(self.__context_p,
                                                          byref(device_p_p))
-        # Instanciate our own libusb_device_p object so we can free
-        # libusb-provided device list. Is this a bug in ctypes that it doesn't
-        # copy pointer value (=pointed memory address) ? At least, it's not so
-        # convenient and forces using such weird code.
-        result = [USBDevice(self, libusb_device_p(x.contents))
-            for x in device_p_p[:device_list_len]]
+        if device_list_len < 0:
+            raise libusb1.USBError(device_list_len)
+        result = []
+        append = result.append
+        for device_p in device_p_p[:device_list_len]:
+            try:
+                # Instanciate our own libusb_device_p object so we can free
+                # libusb-provided device list. Is this a bug in ctypes that it
+                # doesn't copy pointer value (=pointed memory address) ? At
+                # least, it's not so convenient and forces using such weird
+                # code.
+                device = USBDevice(self, libusb_device_p(device_p.contents))
+            except libusb1.USBError, exc:
+                if exc.value != libusb1.LIBUSB_ERROR_ACCESS or \
+                        not skip_on_access_error:
+                    raise
+            else:
+                append(device)
         libusb1.libusb_free_device_list(device_p_p, 0)
         return result
 
-    def openByVendorIDAndProductID(self, vendor_id, product_id):
+    def getByVendorIDAndProductID(self, vendor_id, product_id,
+            skip_on_access_error=False):
+        """
+        Get the first USB device matching given vendor and product ids.
+        Returns an USBDevice instance, or None if no present device match.
+        skip_on_access_error (bool)
+            (see getDeviceList)
+        """
+        for device in self.getDeviceList(
+                skip_on_access_error=skip_on_access_error):
+            if device.getVendorID() == vendor_id and \
+                    device.getProductID() == product_id:
+                result = device
+                break
+        else:
+            result = None
+        return result
+
+    def openByVendorIDAndProductID(self, vendor_id, product_id,
+            skip_on_access_error=False):
         """
         Get the first USB device matching given vendor and product ids.
         Returns an USBDeviceHandle instance, or None if no present device
         match.
+        skip_on_access_error (bool)
+            (see getDeviceList)
         """
-        for device in self.getDeviceList():
-            if device.getVendorID() == vendor_id and \
-                    device.getProductID() == product_id:
-                result = device.open()
-                break
-        else:
-            result = None
+        result = self.getByVendorIDAndProductID(vendor_id, product_id,
+            skip_on_access_error=skip_on_access_error)
+        if result is not None:
+            result = result.open()
         return result
 
     @_validContext
@@ -1644,7 +1834,8 @@ class LibUSBContext(object):
         See libusb_handle_events_locked doc.
         """
         # XXX: does tv parameter need to be exposed ?
-        result = libusb1.libusb_handle_events_locked(self.__context_p, _zero_tv_p)
+        result = libusb1.libusb_handle_events_locked(self.__context_p,
+            _zero_tv_p)
         if result:
             raise libusb1.USBError(result)
 
@@ -1655,5 +1846,14 @@ class LibUSBContext(object):
         """
         return libusb1.libusb_event_handler_active(self.__context_p)
 
-del LibUSBContext._validContext
+del USBContext._validContext
+
+class LibUSBContext(USBContext):
+    """
+    Backward-compatibility alias for USBContext.
+    """
+    def __init__(self):
+        warnings.warn('LibUSBContext is being renamed to USBContext',
+            DeprecationWarning)
+        super(LibUSBContext, self).__init__()
 
